"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.applications.resnet import ResNet50 as ResNet50
from keras.src.applications.resnet import (
    decode_predictions as decode_predictions,
)
from keras.src.applications.resnet import preprocess_input as preprocess_input
